;********************************************************************
;*
;* 
;* 
;* 
;*
;*

;*** Includes
.include "m8def.inc"
.include "lcd.asm"


;*** Constants
.equ	CPUFRQ	=1000000
.equ	CHRLN	=27
.equ	T2SCL	=0x05
.equ	DOT	=0xDB


;*** Ports & Pins
; PORT B
.equ	LCD	=PORTD
.equ	KEYs	=PINC
.equ	KEYMASK	=0x3C


;*** REG defs
.def	null	=r1		; 0
.def	work	=r16		; General working register
.def	temp	=r17		; -"-
.def	cpos	=r23
.def	cnt	=r18		; General counter
.def	dl	=r5
.def	dh	=r6
.def	h	=r22
.def	min	=r21
.def	s	=r20
.def	regs	=r4		; Saves SREG
.def	oldkey	=r2
.def	sb	=r19

.CSEG
;***********************************************************
;* 	INT table
;*

.ORG	0
	rjmp	RESET

.ORG	OVF2addr
	rjmp	T2_OVF


;***********************************************************
;* 	INT routines
;*

T2_OVF:	; Timer 2 overflow
	in	regs,	SREG
	inc	s
	cpi	s,	60
	brlo	EXT2
	clr	s
	inc	min
	cpi	min,	60
	brlo	EXT2
	clr	min
	inc	h
	cpi	h,	24
	brlo	EXT2
	clr	h
	inc	dl
	brne	EXT2
	inc	dh

  EXT2:	out	SREG,	regs
	reti




;***********************************************************
;* 	The program
;*

RESET:	; Start up
	ldi 	work, 	LOW(RAMEND)	; set up stack
	out 	SPL, 	work
	ldi 	work, 	HIGH(RAMEND)
	out 	SPH, 	work					 

	ldi	work,	14
	mov 	h, 	work
	ldi	work,	11
	mov 	min, 	work
	clr 	s

	rcall	INIT_HW

	rcall	READKEYS
	mov	oldkey,	work


MAIN:;	in	work,	KEYs
;	andi	work,	KEYMASK
	rcall	READKEYS
	cp	work,	oldkey
	brne	H_KEY

	ldi	work,	(1<<SM0)|(1<<SM1)|(1<<SE)
	out	MCUCR,	work
	sleep
	nop
	ldi	work,	T2SCL
	out	TCCR2,	work
  w:	in	work,	ASSR
	sbrc	work,	TCN2UB
	  rjmp	w


 RK:	ldi	cpos,	0
	mov	work,	h
	rcall	BIG_OUT
	ldi	cpos,	10
	mov	work,	min
	rcall	BIG_OUT
	ldi	cpos,	20
	mov	work,	s
	rcall	BIG_OUT

	ldi	work,	27+8
	rcall	SET_CURSOR
	ldi	temp,	DOT
	rcall	LCD_DATA
	ldi	work,	3*27+8
	rcall	SET_CURSOR
	ldi	temp,	DOT
	rcall	LCD_DATA
	ldi	work,	27+18
	rcall	SET_CURSOR
	ldi	temp,	DOT
	rcall	LCD_DATA
	ldi	work,	3*27+18
	rcall	SET_CURSOR
	ldi	temp,	DOT
	rcall	LCD_DATA

	rjmp	MAIN


H_KEY:	mov	oldkey,	work
	cpi	work,	KEYMASK
	breq	RH

	sbrs	work,	5
	  inc	h
	sbrs	work,	4
	  dec	h
	sbrs	work,	3
	  inc	min
	sbrs	work,	2
	  dec	min
	sbrs	work,	1
	  clr	s

  RH:	rcall	delay5ms
	rjmp	RK





;***********************************************************
;* 	Additional routines
;*

INIT_HW: ; Inits the micro; no params
	clr 	null

	; Timer setup
	out	TIMSK,	null
	ldi	work,	1<<AS2
	out	ASSR,	work
	ldi	work,	T2SCL
	out	OCR2,	null
	out	TCNT2,	null
	out	TCCR2,	work
  w1:	in	work,	ASSR
	sbrc	work,	TCN2UB
	  rjmp	w1
  w2:	in	work,	ASSR
	sbrc	work,	OCR2UB
	  rjmp	w2
  w3:	in	work,	ASSR
	sbrc	work,	TCR2UB
	  rjmp	w3
	ldi 	work,	(1<<TOIE2)
	out 	TIMSK,	work		; Timer interrupts
	sei
II:
	; PORT setup
	ldi	work,	0xFF
	out	DDRD,	work
	out	PORTD,	work
	out	PORTC,	work
	ldi	work,	0x03
	out	DDRC,	work

	clt
	rcall	LCD_INIT
	rcall	LCD_CLEAR

	clr	work			; load custom chars 
	ldi	ZL,	LOW(c_1<<1)
	ldi	ZH,	HIGH(c_1<<1)
  IH1:	push	work
	rcall	ROM_CHAR
	pop	work
	inc	work
	cpi	work,	8
	brne	IH1

	set
	rcall	LCD_INIT
	rcall	LCD_CLEAR

	clr	work			; load custom chars 
	ldi	ZL,	LOW(c_1<<1)
	ldi	ZH,	HIGH(c_1<<1)
  IH2:	push	work
	rcall	ROM_CHAR
	pop	work
	inc	work
	cpi	work,	8
	brne	IH2

;	ldi	work,	3*27
;	ldi	ZL,	LOW(t_init<<1)
;	ldi	ZH,	HIGH(t_init<<1)
;	rcall	ROM_LINE

	ret





READKEYS:
	sbi	PORTC,	1
	cbi	PORTC,	0
	nop
	in	work,	PINC
	ror	work
	ror	work
	andi	work,	0x0F
	push	work
	cbi	PORTC,	1
	sbi	PORTC,	0
	nop
	in	work,	PINC
	rol	work
	rol	work
	andi	work,	0xF0
	pop	temp
	or	work,	temp
	ret




;***********************************************************
;* 	LCD routines
;*

SET_CURSOR: ; Sets the cursor (DD-RAM address) to 0..31
	; SET_CURSOR (<work> cursorpos 0..31)
	push	work
	cpi	work,	2*CHRLN
	brlo	SC1
	subi	work,	2*CHRLN
	set
	rjmp	SC2
  SC1:	clt
  SC2:	mov	temp,	work		; temp holds the address in DD-RAM of LCD
	cpi	work,	CHRLN		; line 1?
	brlo	SC3
	ldi	temp,	64-CHRLN	; set address for line 1
	add	temp,	work

  SC3:	ori	temp,	0b10000000	; set command bit
	rcall	LCD_COMMAND		; send command "SET DD-RAM ADDRESS"
	pop	work
	ret


LINE_OUT: ; Writes a line to the LCD
	; LINE_OUT (<SRAM @Z> *text)

  LO1:	ldi	cnt,	CHRLN		; write 19 chars
  LO2:	ld	temp,	Z+		; load char from SRAM
	rcall	LCD_DATA		; write it
	dec	cnt			; dec cnt
	brne	LO2			; until cnt = 0
	ret

FIX_OUT: ; Writes a fixed number of chars to the LCD
	; FIX_OUT (<cnt> number of chars, <work> cursorpos, <ROM @Z> *text)
	push	cnt
	push	work
	rcall	SET_CURSOR
  FO:	lpm	temp,	Z+		; load char from SRAM
	rcall	LCD_DATA		; write it
	dec	cnt			; dec cnt
	brne	FO			; until cnt = 0
	pop	work
	pop	cnt
	ret


NUMBER_OUT: ; Writes a decimal number 0..99 at DD-RAM pos
	; NUMBER_OUT (<work> number)
	mov	temp,	work
	clr	cnt

	clr	cnt			; count 10s
  NO3:	cpi	work,	10
	brlo	NO4
	subi	work,	10
	inc	cnt
	rjmp	NO3

  NO4:	ldi	temp,	'0'		; write 10s
	add	temp,	cnt
	push	work
	rcall	LCD_DATA
	pop	work

	ldi	temp,	'0'		; write 1s
	add	temp,	work
	rcall	LCD_DATA

	ret


HEX_OUT:; Writes hex number to LCD
	push	work
	swap	work
	rcall	nibble2hex
	mov	temp,	work
	rcall	LCD_DATA
	pop	work
	rcall	nibble2hex
	mov	temp,	work
	rcall	LCD_DATA
	ret


ROM_LINE: ; Writes a line from ROM
	; ROM_LINE (<work> cursorposition, <ROM @Z> *Text)
	rcall	SET_CURSOR

	ldi	cnt,	CHRLN		; write 24 chars
  RL1:	lpm	temp,	Z+		; load char from ROM
	rcall	LCD_DATA		; write it
	dec	cnt			; dec cnt
	brne	RL1			; until cnt = 0

	ret


ROM_CHAR: ; Writes a custom character from ROM to LCD
	; ROM_CHAR (<work> charnumber 0..7, <ROM @Z> *chardata)

	ldi	temp,	8		; 8
	mul	work,	temp		; * char number

	ldi	cnt,	0		; write 8 pixel lines
  RC1:	mov	temp,	cnt
	add	temp,	r0		; result is max. 56, so only r0 is interesting
	ori	temp,	0b01000000	; set command bit
	rcall	LCD_COMMAND		; send command "SET CG-RAM ADDRESS"
 	lpm	temp,	Z+		; load line from ROM
	rcall	LCD_DATA		; write it
	inc	cnt			; inc cnt
	cpi	cnt,	8
	brne	RC1			; until cnt = 8

	ret


nibble2hex:; work bits 0..3 are converted in "0".."F"
	andi	work,	0x0F
	cpi	work,	10
	brlo	NH1
	subi	work,	249
  NH1:	subi	work,	208
	ret


BIG_OUT: ; Writes a decimal number 0..99 at DD-RAM pos
	; NUMBER_OUT (<work> number, <cpos> cursor pos 0..23)
	mov	temp,	work
	clr	cnt

	clr	cnt			; count 10s
  BO3:	cpi	work,	10
	brlo	BO4
	subi	work,	10
	inc	cnt
	rjmp	BO3

  BO4:	push	work
	mov	temp,	cnt
	mov	work,	cpos
	rcall	BIG_NUMBER
	subi	cpos,	256-4
	pop	work

	mov	temp,	work
	mov	work,	cpos
	rcall	BIG_NUMBER
	subi	cpos,	256-4

	ret


BIG_NUMBER:; temp 0..10 (10=',')
	; work 0..23 (position)
	
	ldi	ZL,	LOW(t_01<<1)
	ldi	ZH,	HIGH(t_01<<1)
	ldi	cnt,	4
	push	work
	push	temp

  LPN:	cpi	temp,	0
	breq	EBN
	dec	temp
	adiw	ZH:ZL,	16
	rjmp	LPN

  EBN:	rcall	FIX_OUT
	subi	work,	256-27
  	rcall	FIX_OUT
	subi	work,	256-27
  	rcall	FIX_OUT
	subi	work,	256-27
  	rcall	FIX_OUT

	pop	temp
	pop	work
	ret	


t_init:
.DB	"LCD 4x27 - Init OK         ",0

t_01:
.DB	3,5,1,' '
t_02:
.DB	7,' ',0,' '
t_03:
.DB	7,' ',0,' '
t_04:
.DB	4,6,2,' '

t_11:
.DB	' ',' ',2,' '
t_12:
.DB	' ',' ',0,' '
t_13:
.DB	' ',' ',0,' '
t_14:
.DB	' ',' ',0,' '

t_21:
.DB	3,5,1,' '
t_22:
.DB	6,6,2,' '
t_23:
.DB	7,' ',' ',' '
t_24:
.DB	4,6,6,' '

t_31:
.DB	3,5,1,' '
t_32:
.DB	' ',6,2,' '
t_33:
.DB	' ',' ',0,' '
t_34:
.DB	4,6,2,' '

t_41:
.DB	7,' ',0,' '
t_42:
.DB	4,6,2,' '
t_43:
.DB	' ',' ',0,' '
t_44:
.DB	' ',' ',0,' '

t_51:
.DB	3,5,5,' '
t_52:
.DB	4,6,6,' '
t_53:
.DB	' ',' ',0,' '
t_54:
.DB	4,6,2,' '

t_61:
.DB	3,5,5,' '
t_62:
.DB	4,6,6,' '
t_63:
.DB	7,' ',0,' '
t_64:
.DB	4,6,2,' '

t_71:
.DB	3,5,1,' '
t_72:
.DB	' ',' ',0,' '
t_73:
.DB	' ',' ',0,' '
t_74:
.DB	' ',' ',0,' '

t_81:
.DB	3,5,1,' '
t_82:
.DB	4,6,2,' '
t_83:
.DB	7,' ',0,' '
t_84:
.DB	4,6,2,' '

t_91:
.DB	3,5,1,' '
t_92:
.DB	4,6,2,' '
t_93:
.DB	' ',' ',0,' '
t_94:
.DB	4,6,2,' '

t_p1:
.DB	"    "
t_p2:
.DB	" ",255,"  "
t_p3:
.DB	"    "
t_p4:
.DB	" ",255,"  "


;.DB	0,1,1,4,5,5,6,7,7,6,"                "
;.DB	0,2,3,0,3,4,0,4,3,4,".               "


c_1: ; custom char 0
.DB	0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07, 0x07
;.DB	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01
c_2: ; custom char 1
.DB	0x1f, 0x1f, 0x1f, 0x07, 0x07, 0x07, 0x07, 0x07
;.DB	0x1F, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x1F
c_3: ; custom char 2
.DB	0x07, 0x07, 0x07, 0x07, 0x07, 0x1f, 0x1f, 0x1f
;.DB	0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1F
c_4: ; custom char 3
.DB	0x1f, 0x1f, 0x1f, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c
;.DB	0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x1F
c_5: ; custom char 4
.DB	0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1f, 0x1f, 0x1f
;.DB	0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x1F
c_6: ; custom char 5
.DB	0x1f, 0x1f, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00
;.DB	0x1F, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x1F
c_7: ; custom char 6
.DB	0x00, 0x00, 0x00, 0x00, 0x00, 0x1f, 0x1f, 0x1f
;.DB	0x1F, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11
c_8: ; custom char 7
.DB	0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c, 0x1c
;.DB	0x1F, 0x11, 0x11, 0x11, 0x11, 0x11, 0x11, 0x1F
;.DB	0x00, 0x07, 0x07, 0x07, 0x07, 0x07, 0x02, 0x04
